import io
import pathlib
import warnings
import zipfile

import numpy as np

from .ws_single import AFMWorkshopFormatWarning, load_csv

__all__ = ["load_map"]


def load_map(path, callback=None, meta_override=None):
    """Load a set of zipped csv AFM workshop data

    If you are recording quantitative force-maps (i.e. multiple
    curves on an x-y-grid) with AFM workshop setups, then you
    might have realized that you get *multiple* .csv files (one
    file per indentation) instead of *one* file that contains all
    the data (as you might be accustomed to from other
    manufacturers). Since afmformats expects one file per
    measurement, it would not be straight forward to obtain a
    properly enumerated quantitative imaging group.

    This function offers a workaround - it loads a zip archive
    created from the the .csv files.

    The files are structured like this::

        Force-Distance Curve
        File Format:    3

        Date:    Wednesday, August 1, 2018
        Time:    1:07:47 PM
        Mode:    Mapping
        Point:    16
        X, um:    27.250000
        Y, um:    27.250000

        Extend Z-Sense(nm),Extend T-B(V),Retract Z-Sense(nm),Retract T-B(V)
        13777.9288,0.6875,14167.9288,1.0917
        13778.9288,0.6874,14166.9288,1.0722
        13779.9288,0.6876,14165.9288,1.0693
        13780.9288,0.6877,14164.9288,1.0824
        13781.9288,0.6875,14163.9288,1.0989
        ...

    Please make sure that the ``Point`` is enumerated from 1
    onwards (and matches the alphanumerical order of the files in
    the archive) and that ``Mode`` is ``Mapping``. The ``X`` and
    ``Y`` coordinates can be used by e.g. PyJibe to display QMap
    data on a grid.


    Parameters
    ----------
    path: str or pathlib.Path
        path to zip file containing AFM workshop .csv files
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    datasets = []
    with zipfile.ZipFile(path) as arc:
        names = sorted(arc.namelist())
        for ii, name in enumerate(names):
            with arc.open(name, "r") as fd:
                tfd = io.TextIOWrapper(fd, encoding="utf-8")
                dd = load_csv(
                    tfd,
                    # recurse into callback with None as default
                    callback=lambda x: callback((ii + x) / len(names))
                    if callback is not None else None,
                    meta_override=meta_override,
                    mode="mapping")
                dd[0]["metadata"]["path"] = pathlib.Path(path)
                cur_enum = dd[0]["metadata"]["enum"]
                if cur_enum != ii + 1:
                    warnings.warn("Dataset 'Point' enumeration mismatch for "
                                  f"'{name}' in '{path}' (expected {ii + 1}, "
                                  f"got {cur_enum})!",
                                  AFMWorkshopFormatWarning)
                datasets += dd
    # Populate missing grid metadata
    xvals = list(set([ad["metadata"]["position x"] for ad in datasets]))
    yvals = list(set([ad["metadata"]["position y"] for ad in datasets]))
    mdgrid = {
        "grid center x": np.mean(xvals),
        "grid center y": np.mean(yvals),
        "grid shape x": len(xvals),
        "grid shape y": len(yvals),
        # grid size in um includes boundaries of pixels
        "grid size x": np.ptp(xvals)*(1 + 1/(len(xvals)-1)),
        "grid size y": np.ptp(yvals)*(1 + 1/(len(yvals)-1)),
    }
    # Update with new metadata (note that grid index x/y is populated via
    # MetaData._autocomplete_grid_metadata)
    [ad["metadata"].update(mdgrid) for ad in datasets]
    return datasets



"""Test of basic opening functionalities"""
import pathlib

import numpy as np

import afmformats
from afmformats.formats.fmt_jpk import load_jpk


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_open_jpk_qi():
    jpkfile = data_path / "fmt-jpk-fd_2020.02.07-16.29.05.036.jpk-qi-data"
    dataset = load_jpk(jpkfile)
    data = dataset[0]["data"]
    force = data["force"]
    height = data["height (measured)"]
    # Verified with visual inspection of force curve in JPK software
    assert np.allclose(force[0], -1.269014596090597e-10)
    assert np.allclose(height[0], 4.716154783699957e-06)


def test_open_jpk_map_enum():
    jpkfile = data_path / "fmt-jpk-fd_2020.02.07-16.29.05.036.jpk-qi-data"
    data = afmformats.load_data(jpkfile)
    # Correct enumeration
    assert data[0].metadata["enum"] == 0
    assert data[1].metadata["enum"] == 1
    assert data[2].metadata["enum"] == 2
    assert data[3].metadata["enum"] == 3


def test_open_jpk_map2():
    jpkfile = data_path / "fmt-jpk-fd_2020.02.07-16.29.05.036.jpk-qi-data"
    dataset = load_jpk(jpkfile)
    assert len(dataset) == 4
    assert dataset[0]["metadata"]["grid index x"] == 0
    assert dataset[1]["metadata"]["grid index x"] == 1
    assert dataset[1]["metadata"]["grid index y"] == 0
    assert dataset[2]["metadata"]["grid index x"] == 2
    assert dataset[3]["metadata"]["grid index x"] == 3
    assert dataset[3]["metadata"]["instrument"] == "JPK01496-H-18-0132"
    assert dataset[3]["metadata"]["software version"] == "6.1.157"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import functools

__all__ = ["LazyData"]


class LazyData(object):
    """Lazily load data from function and kwargs

    The idea is that the experimental data does not have
    to be loaded before the user requests it. Furthermore,
    this reduces the memory footprint (not all data are
    loaded).
    """
    def __init__(self):
        self.loaders = {}

    def __deepcopy__(self, memo):
        # Make sure deepcopy does not copy anything.
        # We might have `io`s, etc.
        return self

    def __contains__(self, key):
        return key in self.loaders

    # cache a few curves so everything is more responsive
    @functools.lru_cache(maxsize=100)
    def __getitem__(self, key):
        if key in self:
            func, kwargs = self.loaders[key]
            return func(**kwargs)

    def __iter__(self):
        for column in self.loaders:
            yield column

    def keys(self):
        return self.loaders.keys()

    def set_lazy_loader(self, column, func, kwargs):
        """Add a lazy loader

        Parameters
        ----------
        column: str
            Column for which to register the loader
        func: callable
            Function to call to get the data
        kwargs:
            Keyword arguments to ``func``
        """
        self.loaders[column] = (func, kwargs)



from .afm_data import AFMData
from .afm_segment import AFMSegment

__all__ = ["AFMForceDistance"]


class AFMForceDistance(AFMData):
    """Base class for AFM force-distance data

    A force-distance dataset consists of an approach and
    a retract curve.
    """
    def __init__(self, *args, **kwargs):
        super(AFMForceDistance, self).__init__(*args, **kwargs)
        #: Dictionary-like interface to the approach segment
        self.appr = AFMSegment(self._raw_data, self._data, segment=0)
        #: Dictionary-like interface to the retract segment
        self.retr = AFMSegment(self._raw_data, self._data, segment=1)

    def __setitem__(self, key, value):
        super(AFMForceDistance, self).__setitem__(key, value)
        if key == "segment":
            # The user changed the segment, which means we have to clear
            # the segment cache.
            self.appr.clear_cache()
            self.retr.clear_cache()

    @property
    def modality(self):
        """Imaging modality"""
        return "force-distance"



from collections import OrderedDict
import copy
import functools
import zipfile

import jprops
import numpy as np

from ... import meta

from . import jpk_data, jpk_meta


__all__ = ["ArchiveCache", "JPKReader"]


class ArchiveCache:
    """Archive cache for fast access to zip data

    If every :class:`JPKReader` has its own instance of `ZipFile`,
    then on macOS (and possibly other OSes), we might run into an
    OSError; [Errno 24] Too many open files
    (https://github.com/AFM-analysis/afmformats/issues/10).
    The problem is, that if we don't leave the `ZipFile`, we
    have to re-open it every time we want to access some data.
    This is a huge overhead.

    The solution is `ArchiveCache`, which keeps a reference to the
    last `max_archives=32` archives and closes the ones that were
    used least.
    """
    open_archives = OrderedDict()
    max_archives = 32

    @staticmethod
    def get(zip_path):
        """Return the (possibly cached) `ZipFile` object for `zip_path`"""
        if zip_path in ArchiveCache.open_archives:
            arc = ArchiveCache.open_archives.pop(zip_path)
        else:
            arc = zipfile.ZipFile(zip_path, mode="r")
        ArchiveCache.open_archives[zip_path] = arc
        # remove any open archives
        too_many = len(ArchiveCache.open_archives) - ArchiveCache.max_archives
        if too_many > 0:
            to_remove = list(ArchiveCache.open_archives.keys())[:too_many]
            for key in to_remove:
                old_arc = ArchiveCache.open_archives.pop(key)
                old_arc.close()
        return arc


class JPKReader(object):
    def __init__(self, path):
        self.path = path

    @functools.lru_cache()
    def __len__(self):
        return len(self.get_index_numbers())

    @property
    @functools.lru_cache()
    def files(self):
        """List of files and folders in the archive"""
        arc = ArchiveCache.get(self.path)
        nlist = arc.namelist()
        maxdigits = int(np.ceil(np.log10(len(nlist)))) + 1
        repstr = "{:0" + "{}".format(maxdigits) + "d}"

        def sortkey(x):
            if x.count("/"):
                xs = x.split("/")
                for ii in range(len(xs)):
                    if xs[ii].isnumeric():
                        xs[ii] = repstr.format(int(xs[ii]))
                return "/".join(xs)
            else:
                return x

        return sorted(nlist, key=sortkey)

    @property
    @functools.lru_cache()
    def hierarchy(self):
        """Format hierarchy ("single" or "indexed")"""
        if "segments/" in self.files:
            return "single"
        elif "index/" in self.files:
            return "indexed"
        else:
            msg = "Cannot determine hierarchy: {}".format(self.path)
            raise NotImplementedError(msg)

    @property
    @functools.lru_cache()
    def _properties_general(self):
        """Return content of "header.properties"""
        arc = ArchiveCache.get(self.path)
        with arc.open("header.properties", "r") as fd:
            props = jprops.load_properties(fd)
        return props

    @property
    @functools.lru_cache()
    def _properties_shared(self):
        """Return content of "shared-data/header.properties"""
        path = "shared-data/header.properties"
        if path in self.files:
            arc = ArchiveCache.get(self.path)
            with arc.open(path, "r") as fd:
                props = jprops.load_properties(fd)
        else:
            props = {}
        return props

    @functools.lru_cache()
    def _get_index_segment_properties(self, index, segment):
        """Return properties fro a specific index and segment

        Parameters
        ----------
        index: int
            Curve index; For "single" hierarchy files, this should be 0.
        segment: int or None
            If None, then no segment-specific properties (e.g.
            approach or retract) are returned.
        """
        # 1. Properties of index
        p_index = self.get_index_path(index) + "header.properties"
        arc = ArchiveCache.get(self.path)
        with arc.open(p_index, "r") as fd:
            prop = jprops.load_properties(fd)

        # 2. Properties of segment (if applicable)
        if segment is not None:
            p_segment = self.get_index_segment_path(index, segment) \
                        + "segment-header.properties"
            with arc.open(p_segment, "r") as fd:
                prop.update(jprops.load_properties(fd))

        # 3. Substitute shared properties
        psprop = self._properties_shared
        # Generate lists of keys and sort them for easier debugging.
        proplist = list(prop.keys())
        proplist.sort()
        pslist = list(psprop.keys())
        pslist.sort()
        # Loop through the segment data and search for lcd-info tags
        for key in proplist:
            # Get line channel data
            if key.count(".*"):
                # Replace the lcd-info tag by the values in the shared
                # properties file:
                # 0, 1, 2, 3, etc.
                pindex = prop[key]
                # lcd-info, force-segment-header-info
                mediator = ".".join(key.split(".")[-2:-1])
                # channel.vDeflection, force-segment-header
                headkey = key.rsplit(".", 2)[0]
                # append a "." here to make sure
                # not to confuse "1" with "10".
                startid = "{}.{}.".format(mediator, pindex)
                for k2 in pslist:
                    if k2.startswith(startid):
                        var = ".".join(k2.split(".")[2:])
                        prop[".".join([headkey, var])] = psprop[k2]

        # 4. Update with general properties
        # (for "single" hierarchy, this coincides with index properties)
        prop.update(self._properties_general)

        # 5. Try to convert numbers to floats and remove NaNs
        for p in list(prop.keys()):
            try:
                prop[p] = float(prop[p])
            except BaseException:
                pass
            else:
                if np.isnan(prop[p]):
                    prop.pop(p)
        return prop

    def get_data(self, column, index, segment=None):
        """Return data for a given column, index, or segment

        Parameters
        ----------
        column: str
            Valid column from :const:`afmformats.afm_data.known_columns`
        index: int
            Curve index in the current archive
        segment: int or None
            Segment index for chosen curve index

        Returns
        -------
        data: 1d ndarray
            Column data
        """
        numsegs = self.get_index_segment_numbers(index)
        if segment is None:
            # Return concatenated data for all segments
            data = []
            for seg in numsegs:
                data.append(self.get_data(column=column, index=index,
                                          segment=seg))
            return np.concatenate(data)
        md = self.get_metadata(index, segment)
        prop = self._get_index_segment_properties(index, segment)
        numsegs = self.get_index_segment_numbers(index)
        # Find the data file that corresponds to the specified column
        if column == "time":
            # get initial time
            start = 0
            if segment != 0:
                for seg in numsegs:
                    if seg < segment:
                        start += self.get_metadata(index, seg)["duration"]

            return np.linspace(start, start + md["duration"],
                               md["point count"], endpoint=False)
        elif column == "segment":
            return np.ones(md["point count"], dtype=np.uint8) * segment
        else:
            # get the segment's data list
            p_seg = self.get_index_segment_path(index, segment)
            loc_list = [ff for ff in self.files if ff.count(p_seg)]
            name, slot, dat = jpk_data.find_column_dat(loc_list, column)
            arc = ArchiveCache.get(self.path)
            with arc.open(dat, "r") as fd:
                data, unit, _ = jpk_data.load_dat_unit(fd, name=name,
                                                       properties=prop,
                                                       slot=slot)
            # verify unit
            if unit != jpk_data.JPK_UNITS[column]:
                raise jpk_data.ReadJPKError("Unknown unit for {}: {}".format(
                    column, unit))
            return data

    @functools.lru_cache()
    def get_index_numbers(self):
        """Return int array with available index numbers

        The numbers is what we refer to as "enum" in afmformats.
        Sometimes individual curves are missing from JPK files.
        These have to be correctly indexed.
        """
        indices = []
        if self.hierarchy == "single":
            indices.append(0)
        else:
            # TODO: is there a more efficient way?
            for ff in self.files:
                if (ff.startswith("index/")
                        and ff.count("/") == 2
                        and ff.endswith("/")):
                    indices.append(int(ff.split("/")[1]))
        indices = np.array(indices, dtype=int)
        return indices

    @functools.lru_cache()
    def get_index_path(self, index):
        """Return the path in the zip file for a specific curve index"""
        enum = self.get_index_numbers()[index]
        if self.hierarchy == "single":
            path = ""
        elif self.hierarchy == "indexed":
            path = "index/{}/".format(enum)
        else:
            raise NotImplementedError("No rule to get path for hierarchy "
                                      + "'{}'!".format(self.hierarchy))
        if path and path not in self.files:
            raise IndexError("Cannot find path for index '{}' ".format(index)
                             + " (enum '{}')!".format(enum))
        return path

    @functools.lru_cache()
    def get_index_segment_numbers(self, index):
        """Return available segment numbers for an index"""
        segments = []
        seg = 0
        while True:
            try:
                self.get_index_segment_path(index, seg)
            except IndexError:
                break
            else:
                segments.append(seg)
                seg += 1
        return segments

    @functools.lru_cache()
    def get_index_segment_path(self, index, segment):
        """Return the path in the zip file for a specific index and segment"""
        enum = self.get_index_numbers()[index]
        if self.hierarchy == "single":
            path = "segments/{}/".format(segment)
        elif self.hierarchy == "indexed":
            path = "index/{}/segments/{}/".format(enum, segment)
        else:
            raise NotImplementedError("No rule to get path for hierarchy "
                                      + "'{}'!".format(self.hierarchy))
        if path not in self.files:
            raise IndexError("Cannot find path for index '{}' ".format(index)
                             + "(enum '{}')".format(enum))
        return path

    @functools.lru_cache()
    def get_metadata(self, index, segment=None):
        """Return the metadata for a specific index and segment

        Parameters
        ----------
        index: int
            Curve index; For "single" hierarchy files, this should be 0.
        segment: int or None
            If None, then all segment-specific properties (e.g.
            approach and retract) are returned.
        """
        if segment is None:
            md = meta.MetaData()
            # reverse order, because we want "time" from first md
            for seg in self.get_index_segment_numbers(index)[::-1]:
                mdap = copy.deepcopy(self.get_metadata(index, seg))
                if "duration" in md:
                    md["duration"] += mdap.pop("duration")
                if "point count" in md:
                    md["point count"] += mdap.pop("point count")
                md.update(mdap)
            return md
        prop = self._get_index_segment_properties(index=index, segment=segment)
        # 1. Populate with primary metadata
        md = meta.MetaData()
        recipe = jpk_meta.get_primary_meta_recipe()
        for key in recipe:
            for vari in recipe[key]:
                if vari in prop:
                    md[key] = prop[vari]
                    break
        for mkey in ["spring constant",
                     "sensitivity",
                     ]:
            if mkey not in md:
                msg = "Missing meta data: '{}'".format(mkey)
                raise jpk_meta.ReadJPKMetaKeyError(msg)

        md["software"] = "JPK"

        md["enum"] = int(self.get_index_numbers()[index])
        md["path"] = self.path

        # 2. Populate with secondary metadata
        recipe_2 = jpk_meta.get_secondary_meta_recipe()
        md_im = {}

        for key in recipe_2:
            for vari in recipe_2[key]:
                if vari in prop:
                    md_im[key] = prop[vari]
                    break

        curve_type = md_im["curve type"]
        curve_conv = {"extend": "approach",
                      "pause": "intermediate",
                      "retract": "retract"}
        curseg = curve_conv[curve_type]

        if curseg in ["approach", "retract"]:
            md[f"rate {curseg}"] = md["point count"] / \
                                   md_im["segment duration"]
            zrange = abs(md_im["z start"] - md_im["z end"])
            md[f"speed {curseg}"] = zrange / md_im["segment duration"]
        md[f"duration {curseg}"] = md_im["segment duration"]

        num_segments = len(self.get_index_segment_numbers(0))
        if num_segments == 2:
            md["imaging mode"] = "force-distance"
        elif num_segments == 3:
            if segment == 1:
                # for creep-compliance and stress-relaxation, we get the
                # extra information from segment 1.
                if md_im["curve type"] != "pause":
                    raise ValueError("Segment 1 must be of type 'pause'!")
                pause_type = md_im["segment pause type"]
                if pause_type == "constant-force-pause":
                    md["imaging mode"] = "creep-compliance"
                elif pause_type == "constant-height-pause":  # not sure
                    md["imaging mode"] = "stress-relaxation"
                else:
                    raise ValueError(f"Unexprected pause type '{pause_type}'!")
        else:
            raise ValueError(f"Unexpected number of segments: {num_segments}!")

        md["curve id"] = "{}:{:g}".format(md["session id"],
                                          md_im["position index"])
        if "setpoint [V]" in md_im:
            md["setpoint"] = md_im["setpoint [V]"] * \
                             md["spring constant"] * md["sensitivity"]

        # date and time
        if curseg == "approach":
            md["date"], md["time"], _ = md_im["time stamp"].split()

        # Convert designated keys to integers
        integer_keys = ["grid shape x",
                        "grid shape y",
                        "grid index x",
                        "grid index y",
                        "point count",
                        ]
        for ik in integer_keys:
            if ik in md:
                md[ik] = int(round(md[ik]))
        return md



"""Test AFMFormatRecipe"""
import pathlib
import shutil
import tempfile

import pytest

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_bad_recipe_loader_missing():
    recipe = {
        "descr": "unknown description",
        "maker": "unknown maker",
        "modality": "force-distance",
        "suffix": ".unknown",
    }
    with pytest.raises(ValueError):
        afmformats.formats.register_format(recipe)


def test_bad_recipe_loader_not_callable():
    recipe = {
        "descr": "unknown description",
        "loader": "peter",
        "maker": "unknown maker",
        "modality": "force-distance",
        "suffix": ".unknown",
    }
    with pytest.raises(ValueError):
        afmformats.formats.register_format(recipe)


def test_bad_recipe_modality_invlaid():
    recipe = {
        "descr": "unknown description",
        "loader": lambda x: x,
        "maker": "unknown maker",
        "suffix": ".unknown",
        "modality": "invalid",
    }
    with pytest.raises(ValueError):
        afmformats.formats.register_format(recipe)


def test_bad_recipe_modality_missing():
    recipe = {
        "descr": "unknown description",
        "loader": lambda x: x,
        "maker": "unknown maker",
        "suffix": ".unknown",
    }
    with pytest.raises(ValueError):
        afmformats.formats.register_format(recipe)


def test_bad_recipe_suffix_invalid():
    recipe = {
        "descr": "unknown description",
        "loader": lambda x: x,
        "maker": "unknown maker",
        "modality": "force-distance",
    }
    with pytest.raises(ValueError):
        afmformats.formats.register_format(recipe)


@pytest.mark.parametrize("name, is_valid",
    [("fmt-jpk-fd_spot3-0192.jpk-force", True),  # noqa: E128
     ("fmt-jpk-fd_map2x2_extracted.jpk-force-map", True),
     ("fmt-jpk-cl_calibration_force-save-2015.02.04-11.25.21.294.jpk-force",
      False),
     ])
def test_find_data(name, is_valid):
    file = data_path / name
    file_list = afmformats.find_data(file)
    assert bool(len(file_list)) == is_valid


def test_find_data_recursively():
    td = pathlib.Path(tempfile.mkdtemp(prefix="find_data_recursively_"))
    td2 = td / "test" / "recur"
    td2.mkdir(exist_ok=True, parents=True)
    td3 = td / "test" / "recur2"
    td3.mkdir(exist_ok=True, parents=True)
    shutil.copy2(data_path / "fmt-jpk-fd_spot3-0192.jpk-force", td2)
    shutil.copy2(
        data_path /
        "fmt-jpk-cl_calibration_force-save-2015.02.04-11.25.21.294.jpk-force",
        td3)
    file_list = afmformats.find_data(td)
    assert len(file_list) == 1
    assert file_list[0].samefile(td2 / "fmt-jpk-fd_spot3-0192.jpk-force")


def test_find_data_invalid_missing():
    td = pathlib.Path(tempfile.mkdtemp(prefix="find_data_invalid_"))
    shutil.copy2(data_path / "fmt-jpk-fd_spot3-0192.jpk-force",
                 td / "spot.unknown")
    file_list = afmformats.find_data(td)
    assert not file_list



import pathlib

from igor2 import binarywave
import numpy as np

__all__ = ["load_igor"]


def load_igor(path, callback=None, meta_override=None):
    """Load Asylum Research (Igor) binarywave .ibw files

    The raw data are loaded with the Python module "igor"
    (http://blog.tremily.us/posts/igor/).
    The way column labels are assigend to the data is kind of hacky.
    The metadata assignment is largely guessed.

    Test data were provided by Nicolas Hauck :cite:`Hauck2018`.

    Parameters
    ----------
    path: str or pathlib.Path
        path to in .ibw data file
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    if meta_override is None:
        meta_override = {}
    else:
        # just make sure nobody expects a different result for the forces
        for key in ["sensitivity", "spring constant"]:
            if key in meta_override:
                raise NotImplementedError(
                    f"Setting metadata such as '{key}' is not implemented!")

    ibw = binarywave.load(path)
    wdata = ibw["wave"]["wData"]
    notes = {}
    for line in str(ibw["wave"]["note"]).split("\\r"):
        if line.count(":"):
            key, val = line.split(":", 1)
            notes[key] = val.strip()

    # Metadata
    metadata = {}
    # acquisition
    metadata["imaging mode"] = "force-distance"
    metadata["feedback mode"] = notes["ImagingMode"].lower()
    metadata["rate approach"] = float(notes["NumPtsPerSec"])
    metadata["rate retract"] = float(notes["NumPtsPerSec"])
    metadata["sensitivity"] = float(notes["InvOLS"])
    if notes["TriggerChannel"] == "Force":
        metadata["setpoint"] = float(notes["TriggerPoint"])
    metadata["spring constant"] = float(notes["SpringConstant"])

    # dataset
    metadata["duration"] = float(notes["InputTime"])
    metadata["enum"] = 1
    metadata["speed approach"] = float(notes["ApproachVelocity"])
    metadata["speed retract"] = float(notes["RetractVelocity"])

    # setup
    metadata["instrument"] = notes["MicroscopeModel"]
    metadata["software"] = "Asylum Research"
    metadata["software version"] = notes["Version"]

    # storage
    metadata["date"] = notes["Date"]
    metadata["path"] = pathlib.Path(path)
    metadata["point count"] = wdata.shape[0]
    metadata["time"] = notes["Time"]

    # Data
    labels = []
    # deal with something like this:
    # 'labels': [[], [b'', b'Raw', b'Defl', b'ZSnsr'], [], []],
    for ll in ibw["wave"]["labels"]:
        for li in ll:
            if li:
                labels.append(li.decode())
    assert len(labels) == wdata.shape[1]

    data = {}
    for pkey in ["Raw", "Height"]:
        if pkey in labels:
            # height is in [m]
            data["height (piezo)"] = -wdata[:, labels.index(pkey)]
            break

    for mkey in ["ZSnsr", "ZSensor"]:
        if mkey in labels:
            # height is in [m]
            data["height (measured)"] = -wdata[:, labels.index(mkey)]
            break

    for fkey in ["Defl", "Deflection"]:
        if fkey in labels:
            # force is in [m] (convert to [N])
            data["force"] = wdata[:, labels.index(fkey)] \
                * metadata["spring constant"]
            break

    data["segment"] = np.zeros(wdata.shape[0], dtype=np.uint8)

    # missing metadata
    metadata["z range"] = np.ptp(data["height (piezo)"])
    metadata.update(meta_override)

    dataset = [{"data": data,
                "metadata": metadata,
                }]

    if callback is not None:
        callback(1)

    return dataset


recipe_ibw = {
    "descr": "binarywave",
    "loader": load_igor,
    "suffix": ".ibw",
    "modalities": ["force-distance"],
    "maker": "Asylum Research",
}



def fbool(value):
    """boolean"""
    if isinstance(value, str):
        value = value.lower()
        if value == "false":
            value = False
        elif value == "true":
            value = True
        else:
            value = bool(float(value))
    else:
        value = bool(float(value))
    return value


def fint(value):
    """integer"""
    if isinstance(value, str):
        # strings might have been saved wrongly as booleans
        value = value.lower()
        if value == "false":
            value = 0
        elif value == "true":
            value = 1
        else:
            value = int(float(value))
    else:
        value = int(float(value))
    return value


def vd_str_in(alist):
    """Return a validator that tests whether a string is in a list"""
    def str_in(value):
        if value not in alist:
            raise ValueError("Invalid value '{}', must be in {}!".format(
                value, alist))
        return value
    return str_in


__all__ = [e for e in dir() if not e.startswith("__")]



"""Test NT-MDT text export format"""
import pathlib
import tempfile

import numpy as np
import pytest

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_detect():
    path = data_path / ("fmt-ntmdt-txt-fd_2015_01_17_gel4-0,1_mQ_adh"
                        + "_6B_Curve_DFL_Height_51.txt")
    recipe = afmformats.formats.get_recipe(path)
    assert recipe.descr == "exported by NT-MDT Nova"


def test_detect_bad():
    path = data_path / ("fmt-ntmdt-txt-fd_2015_01_17_gel4-0,1_mQ_adh"
                        + "_6B_Curve_DFL_Height_51.txt")
    _, tf = tempfile.mkstemp(suffix=".txt", prefix="afmformats_test")
    data = pathlib.Path(path).read_text()
    data.replace("\t", ",")
    with pathlib.Path(tf).open("w") as fd:
        # add a header line without any hashes
        fd.write("header without hash")
        fd.write(data)

    with pytest.raises(afmformats.errors.FileFormatNotSupportedError):
        afmformats.formats.get_recipe(tf)


def test_open_simple():
    path = data_path / ("fmt-ntmdt-txt-fd_2015_01_17_gel4-0,1_mQ_adh"
                        + "_6B_Curve_DFL_Height_51.txt")
    meta_override = {"spring constant": 0.055, "sensitivity": 61}
    data = afmformats.load_data(path=path, meta_override=meta_override)[0]
    assert data["force"].size == 1644
    # maximum force is at end of 1st segment
    assert np.sum(data["segment"] == 0) == np.argmax(data["force"])+1


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import functools

import numpy as np

from .afm_group import AFMGroup
from .meta import META_FIELDS


__all__ = ["AFMQMap", "qmap_feature", "unit_scales"]


def qmap_feature(name, unit, cache=False):
    """Decorator for labeling AFMQMap features

    The name and unit are stored as properties of the wrapped function.
    In addition, the return value of the function can be cached (see
    `cache` argument).

    Parameters
    ----------
    name: str
        Name of the feature
    unit: str
        Unit of the returned feature
    cache: bool or callable
        If boolean, determines whether the feature data should be
        cached or not. If callable, the callable gets an instance
        of AFMData as an argument and should return an identifier
        (str) for the current value. If that identifier is the
        same as in the cache, then the cached value is used.
    """
    def attribute_setter(func):
        """Decorator that sets the necessary attributes

        The outer decorator is used to obtain the attributes.
        This inner decorator returns the actual function that
        wraps the feature function.
        """
        func.name = name
        func.unit = unit
        if isinstance(cache, bool):
            if cache:
                # cache everything once
                func.cache_mode = "static"
                func.cache_values = {}
            else:
                # disable caching
                func.cache_mode = "disabled"
        else:
            assert isinstance(cache, callable)
            func.cache_mode = "variable"
            func.cache_values = {}
            func.cache_ids = {}
            func.cache_getid = cache

        @functools.wraps(func)
        def cached_func(afmdata):
            afmdataid = hex(id(afmdata))
            if func.cache_mode == "disabled":
                # no caching
                return func(afmdata)
            elif func.cache_mode == "static":
                # caching is done only once
                if afmdataid not in func.cache_values:
                    func.cache_values[afmdataid] = func(afmdata)
                return func.cache_values[afmdataid]
            else:
                # we have to check whether we have to recompute
                cid = func.cache_getid(afmdata)
                assert cid, "Must not be empty string"
                if func.cache_ids.get(afmdataid, "") != cid:
                    # recompute the value
                    func.cache_values[afmdataid] = func(afmdata)
                    func.cache_ids[afmdataid] = cid
                return func.cache_values[afmdataid]

        return cached_func

    return attribute_setter


class AFMQMap:
    """Management of quantitative AFM data on a grid"""
    def __init__(self, path_or_group, meta_override=None, callback=None,
                 modality=None, data_classes_by_modality=None):
        """
        Parameters
        ----------
        path_or_group: str or pathlib.Path or afmformats.afm_group.AFMGroup
            The path to the data file or an instance of `AFMGroup`
        meta_override: dict
            Dictionary with metadata that is used when loading the data
            in `path`.
        callback: callable or None
            A method that accepts a float between 0 and 1
            to externally track the process of loading the data.
        data_classes_by_modality: dict
            Override the default AFMData class to use for managing the data
            (see :data:`default_data_classes_by_modality`): This is e.g.
            used by :ref:`nanite:index` to pass `Indentation` (which is a
            subclass of the default `AFMForceDistance`) for handling
            "force-indentation" data.
        """
        if isinstance(path_or_group, AFMGroup):
            group = path_or_group
            if meta_override is not None:
                raise ValueError(
                    "Specifying `meta_override` for an AFMGroup instance "
                    "that is already populated is meaningless.")
        else:
            group = AFMGroup(path=path_or_group,
                             meta_override=meta_override,
                             callback=callback,
                             modality=modality,
                             data_classes_by_modality=data_classes_by_modality)
        #: AFM data (instance of :class:`afmformats.afm_group.AFMGroup`)
        self.group = group

        # sanity check (make sure that all necessary metadata are available)
        missing_keys = []
        for key in META_FIELDS["qmap"]:
            if key not in self.group[0].metadata:
                missing_keys.append(key)
        if missing_keys:
            raise ValueError(f"QMap metadata missing for '{self.group.path}': "
                             + ", ".join(missing_keys))

        # Register feature functions
        #: Available features
        self.features = []
        self._feature_funcs = {}
        self._feature_units = {}
        fnames = [f for f in dir(self) if f.startswith("feat_")]
        for fn in fnames:
            func = getattr(self, fn)
            assert func.name not in self._feature_funcs
            assert func.name not in self._feature_units
            self._feature_funcs[func.name] = func
            self._feature_units[func.name] = func.unit
            self.features.append(func.name)

    def _map_grid(self, coords, map_data):
        """Create a 2D map from 1D coordinates and data

        The .jpk-force-map file format stores the map data in a
        seemingly arbitrary way. This method converts a set of
        coordinates and map data values to a 2D map.

        Parameters
        ----------
        coords: list-like (length N) with tuple of ints
            The x- and y-coordinates [px].
        map_data: list-like (length N)
            The data to be mapped.

        Returns
        -------
        x, y: 1d ndarrays
            The x- and y-values that label the axes of the map
        map2d: 2d ndarray
            The ordered map data.
        """
        shape = self.shape
        extent = self.extent

        coords = np.array(coords)
        map_data = np.array(map_data)

        xn, yn = int(shape[0]), int(shape[1])

        # Axes ticks
        x, dx = np.linspace(extent[0], extent[1], xn,
                            endpoint=False, retstep=True)
        y, dy = np.linspace(extent[2], extent[3], yn,
                            endpoint=False, retstep=True)
        x += dx/2
        y += dy/2

        # Output map
        map2d = np.zeros((yn, xn), dtype=float)*np.nan
        for ii in range(map_data.shape[0]):
            # Determine the coordinate in the output array
            xi, yi = coords[ii]
            # Write to the output array
            map2d[yi, xi] = map_data[ii]

        return x, y, map2d

    @property
    @functools.lru_cache(maxsize=32)
    def extent(self):
        """extent (x1, x2, y1, y2) [µm]"""
        afmdata0 = self.group[0]
        # get extent of the map
        sx = afmdata0.metadata["grid size x"] * 1e6
        sy = afmdata0.metadata["grid size y"] * 1e6
        cx = afmdata0.metadata["grid center x"] * 1e6
        cy = afmdata0.metadata["grid center y"] * 1e6
        extent = (cx - sx/2, cx + sx/2,
                  cy - sy/2, cy + sy/2,
                  )
        return extent

    @property
    @functools.lru_cache(maxsize=32)
    def shape(self):
        """shape of the map [px]"""
        afmdata0 = self.group[0]
        # get shape of the map
        shape = (afmdata0.metadata["grid shape x"],
                 afmdata0.metadata["grid shape y"]
                 )
        return shape

    @staticmethod
    @qmap_feature(name="data: height base point",
                  unit="µm",
                  cache=True)
    def feat_core_data_height_base_point_um(afmdata):
        """Compute the lowest height (measured)"""
        height = np.min(afmdata["height (measured)"])
        value = height / unit_scales["µ"]
        return value

    @staticmethod
    @qmap_feature(name="data: piezo range",
                  unit="µm",
                  cache=True)
    def feat_core_data_piezo_range_um(afmdata):
        """Compute peak-to-peak piezo range"""
        return afmdata.metadata["z range"] / unit_scales["µ"]

    @staticmethod
    @qmap_feature(name="data: scan order",
                  unit="",
                  cache=True)
    def feat_core_data_scan_order(afmdata):
        """Return the enumeration of the dataset"""
        return afmdata.enum

    @functools.lru_cache(maxsize=32)
    def get_coords(self, which="px"):
        """Get the qmap coordinates for each curve in `AFMQMap.group`

        Parameters
        ----------
        which: str
            "px" for pixels or "um" for microns.
        """
        if which not in ["px", "um"]:
            raise ValueError("`which` must be 'px' or 'um'!")

        if which == "px":
            kx = "grid index x"
            ky = "grid index y"
            mult = 1
        else:
            kx = "position x"
            ky = "position y"
            mult = 1e6
        coords = []
        for afmdata in self.group:
            # We assume that kx and ky are given. This has to be
            # ensured by the file format reader for qmaps.
            cc = [afmdata.metadata[kx] * mult, afmdata.metadata[ky] * mult]
            coords.append(cc)
        return np.array(coords)

    def get_qmap(self, feature, qmap_only=False):
        """Return the quantitative map for a feature

        Parameters
        ----------
        feature: str
            Feature to compute map for (see :data:`QMap.features`)
        qmap_only:
            Only return the quantitative map data,
            not the coordinates

        Returns
        -------
        x, y: 1d ndarray
            Only returned if `qmap_only` is False; Pixel grid
            coordinates along x and y
        qmap: 2d ndarray
            Quantitative map
        """
        coords = self.get_coords(which="px")

        map_data = []
        ffunc = self._feature_funcs[feature]
        for afmdata in self.group:
            val = ffunc(afmdata)
            map_data.append(val)

        x, y, qmap = self._map_grid(coords=coords, map_data=map_data)
        if qmap_only:
            return qmap
        else:
            return x, y, qmap


#: Scale conversion helper
unit_scales = {
    "": 1,
    "k": 1e3,
    "m": 1e-3,
    "µ": 1e-6,
    "n": 1e-9,
    "p": 1e-12
    }



import pytest

from afmformats import parse_funcs


def assert_false(value):
    assert isinstance(value, bool)
    assert not value


def assert_true(value):
    assert isinstance(value, bool)
    assert value


def test_fbool():
    assert_true(parse_funcs.fbool("true"))
    assert_true(parse_funcs.fbool("True"))
    assert_true(parse_funcs.fbool(True))
    assert_true(parse_funcs.fbool(1))
    assert_true(parse_funcs.fbool(1.0))

    assert_false(parse_funcs.fbool("false"))
    assert_false(parse_funcs.fbool("False"))
    assert_false(parse_funcs.fbool(False))
    assert_false(parse_funcs.fbool(0))
    assert_false(parse_funcs.fbool(0.0))

    with pytest.raises(ValueError):
        parse_funcs.fbool("peter")


def test_fint():
    assert parse_funcs.fint("10") == 10
    assert parse_funcs.fint("-11") == -11
    assert parse_funcs.fint("True") == 1
    assert parse_funcs.fint("true") == 1
    assert parse_funcs.fint("false") == 0
    assert parse_funcs.fint("False") == 0

    with pytest.raises(ValueError):
        parse_funcs.fint("peter")



"""Test .tab format functionalities"""
import pathlib
import tempfile

import numpy as np

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_open_0_13_3():
    fdat = afmformats.load_data(data_path / "fmt-hdf5-fd_version_0.13.3.h5")[0]
    assert fdat.metadata["imaging mode"] == "force-distance"
    assert fdat["force"][2000] == -7.038311459845539e-10


def test_hdf5_metadata_contains():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="force-distance")[0]
    _, path = tempfile.mkstemp(suffix=".h5", prefix="afmformats_test_")
    fdat.export_data(path, metadata=True, fmt="hdf5")
    fdat2 = afmformats.load_data(path, modality="force-distance")[0]
    assert "force" in fdat2


def test_save_open_with_metadata():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="force-distance")[0]
    _, path = tempfile.mkstemp(suffix=".h5", prefix="afmformats_test_")
    fdat.export_data(path, metadata=True, fmt="hdf5")
    fdat2 = afmformats.load_data(path, modality="force-distance")[0]
    # compare metadata
    for key in fdat.metadata:
        if key in ["path", "format"]:
            assert fdat.metadata[key] != fdat2.metadata[key]
        else:
            assert fdat.metadata[key] == fdat2.metadata[key]
    for col in fdat.columns:
        assert np.allclose(fdat[col], fdat2[col], atol=0)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



import pathlib

from .. import errors
from .. import meta
from .fmt_hdf5 import recipe_hdf5
from .fmt_igor import recipe_ibw
from .fmt_jpk import (
    recipe_jpk_force,
    recipe_jpk_force_map,
    recipe_jpk_force_qi_data,
    recipe_jpk_force_qi_series,
)
from .fmt_tab import recipe_tab
from .fmt_ntmdt_txt import recipe_ntmdt_txt
from .fmt_workshop import recipe_workshop_single, recipe_workshop_map
from ..mod_force_distance import AFMForceDistance
from ..mod_creep_compliance import AFMCreepCompliance


__all__ = ["AFMFormatRecipe", "find_data", "get_recipe", "load_data",
           "default_data_classes_by_modality", "formats_available",
           "formats_by_suffix", "formats_by_modality", "supported_extensions"]


class AFMFormatRecipe(object):
    def __init__(self, recipe):
        """A wrapper class for file format recipes

        Parameters
        ----------
        recipe: dict
            file format recipe
        """
        self.recipe = recipe

        # check loader
        if not callable(self.loader):
            raise ValueError(
                "'loader' must be callable: '{}'".format(self.loader))
        self.origin = self.loader.__module__

        # check modality
        if not set(self.modalities) <= set(meta.IMAGING_MODALITIES):
            raise ValueError("'modalities' must be in {}, got '{}'!".format(
                meta.IMAGING_MODALITIES, self.modalities))

        # check detect
        if "detect" in self.recipe:
            if not callable(self.recipe["detect"]):
                raise ValueError(
                    "'detect' must be callable: '{}'".format(
                        self.recipe["detect"]))

    def __getitem__(self, key):
        # backwards compatibility
        return getattr(self, key)

    def __repr__(self):
        repre = "<{} from '{}' at {}>".format(self.__class__.__name__,
                                              self.origin,
                                              hex(id(self)))
        return repre

    @property
    def descr(self):
        """description of file format"""
        return self.recipe.get("descr", "no description")

    @property
    def loader(self):
        """method for loading the data"""
        if "loader" in self.recipe:
            return self.recipe["loader"]
        else:
            raise ValueError("No loader defined!")

    @property
    def maker(self):
        """who introduced the file format"""
        return self.recipe.get("maker", "unknown maker")

    @property
    def modalities(self):
        """list of supported AFM imaging modalities"""
        if "modalities" in self.recipe:
            return self.recipe["modalities"]
        else:
            raise ValueError(f"No modalities defined for recipe {self}!")

    @property
    def suffix(self):
        """file format suffix"""
        if "suffix" in self.recipe:
            return self.recipe["suffix"]
        else:
            raise ValueError("No suffix defined for recipe {}!".format(self))

    def detect(self, path):
        """Determine whether `path` can be opened with this recipe

        Returns
        -------
        valid: bool
            True if `path` is openable, False otherwise.

        Notes
        -----
        If the underlying recipe does not implement a "detect"
        function, then only the file extension is checked.
        """
        valid = False
        path = pathlib.Path(path)
        # basic check for suffix
        try:
            if path.suffix == self.suffix:
                valid = True
        except ValueError:
            pass

        # advanced check with "detect"
        if valid and "detect" in self.recipe:
            fdetect = self.recipe["detect"]
            valid = fdetect(path)

        return valid

    def get_modality(self, path):
        """Determine modality of a path

        If a recipe provides several modalities, load the
        dataset and get the modality from the metadata.
        """
        if len(self.modalities) == 1:
            modality = self.modalities[0]
        else:
            metadata = self.loader(path)[0]["metadata"]
            modality = metadata["imaging mode"]
        return modality


def find_data(path, modality=None):
    """Recursively find valid AFM data files

    Parameters
    ----------
    path: str or pathlib.Path
        file or directory
    modality: str
        modality of the measurement ("force-distance")

    Returns
    -------
    file_list: list of pathlib.Path
        list of valid AFM data files
    """
    path = pathlib.Path(path)
    file_list = []
    if path.is_dir():
        # recurse into subdirectories
        for pp in path.rglob("*"):
            if pp.is_file():
                file_list += find_data(pp, modality=modality)
    else:
        try:
            get_recipe(path=path, modality=modality)
        except errors.FileFormatNotSupportedError:
            # not a valid file format
            pass
        else:
            # valid file format
            file_list.append(path)
    return file_list


def get_recipe(path, modality=None):
    """Return the file format recipe for a given path

    Parameters
    ----------
    path: str or pathlib.Path
        file or directory
    modality: str
        modality of the measurement (see :const:`IMAGING_MODALITIES`)

    Returns
    -------
    recipe: AFMFormatRecipe
        file format recipe
    """
    path = pathlib.Path(path)

    if path.suffix not in formats_by_suffix:
        raise errors.FileFormatNotSupportedError(
            f"No recipe for suffix '{path.suffix}' (file '{path}')!")

    recipes = formats_by_suffix[path.suffix]
    for rec in recipes:
        if ((modality is None or modality in rec.modalities)
                and rec.detect(path)):
            break
    else:
        raise errors.FileFormatNotSupportedError(
            f"Could not determine file format recipe for '{path}'!")

    return rec


def load_data(path, meta_override=None, modality=None,
              data_classes_by_modality=None, diskcache=False,
              callback=None):
    """Load AFM data

    Parameters
    ----------
    path: str or pathlib.Path
        Path to AFM data file
    meta_override: dict
        Metadata dictionary that overrides experimental metadata
    modality: str
        Which acquisition modality to use (e.g. "force-distance")
    data_classes_by_modality: dict
        Override the default AFMData class to use for managing the data
        (see :data:`default_data_classes_by_modality`): This is e.g.
        used by :ref:`nanite:index` to pass `Indentation` (which is a
        subclass of the default `AFMForceDistance`) for handling
        "force-indentation" data.
    diskcache: bool
        Whether to use caching (not implemented)
    callback: callable
        A method that accepts a float between 0 and 1
        to externally track the process of loading the data

    Returns
    -------
    afm_list: list of afmformats.afm_data.AFMData
        List where each element is on AFMData curve
    """
    if meta_override is None:
        meta_override = {}
    if data_classes_by_modality is None:
        data_classes_by_modality = {}
    path = pathlib.Path(path)
    if path.suffix in formats_by_suffix:
        afmdata = []
        cur_recipe = get_recipe(path, modality=modality)
        loader = cur_recipe.loader
        if modality is None:
            modality = cur_recipe.get_modality(path)
            fix_modality = False
        else:
            fix_modality = True
        if modality in data_classes_by_modality:
            afm_data_class = data_classes_by_modality[modality]
        else:
            afm_data_class = default_data_classes_by_modality[modality]
        for dd in loader(path, callback=callback,
                         meta_override=meta_override):
            dd["metadata"]["format"] = "{} ({})".format(cur_recipe["maker"],
                                                        cur_recipe["descr"])
            if fix_modality and dd["metadata"]["imaging mode"] != modality:
                # The user explicitly requested this modality.
                continue
            ddi = afm_data_class(data=dd["data"],
                                 metadata=dd["metadata"],
                                 diskcache=diskcache)
            afmdata.append(ddi)
    else:
        raise ValueError("Unsupported file extension: '{}'!".format(path))
    return afmdata


def register_format(recipe):
    """Registers a file format from a recipe dictionary"""
    afr = AFMFormatRecipe(recipe)
    formats_available.append(afr)
    # suffix
    if afr.suffix not in formats_by_suffix:
        formats_by_suffix[afr.suffix] = []
    formats_by_suffix[afr.suffix].append(afr)
    # modality
    for modality in afr.modalities:
        if modality not in formats_by_modality:
            formats_by_modality[modality] = []
        formats_by_modality[modality].append(afr)
    # supported extensions
    if afr.suffix not in supported_extensions:  # avoid duplicates
        supported_extensions.append(afr.suffix)
    supported_extensions.sort()


#: dictionary with default data classes for each modality
default_data_classes_by_modality = {
    "force-distance": AFMForceDistance,
    "creep-compliance": AFMCreepCompliance,
}

#: available/supported file formats
formats_available = []

#: available file formats in a dictionary with suffix keys
formats_by_suffix = {}

#: available file formats in a dictionary for each modality
formats_by_modality = {}

#: list of supported extensions
supported_extensions = []

for _recipe in [
    recipe_hdf5,
    recipe_ibw,
    recipe_jpk_force,
    recipe_jpk_force_map,
    recipe_jpk_force_qi_data,
    recipe_jpk_force_qi_series,
    recipe_ntmdt_txt,
    recipe_tab,
    recipe_workshop_map,
    recipe_workshop_single,
]:
    register_format(_recipe)



import pathlib
import tempfile

import h5py
import numpy as np

import afmformats


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_export_hdf5():
    """Test for hotfixes in 0.7.1"""
    jpkfile = data_path / "fmt-jpk-cc_pr14-brain-2021.06.30.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="creep-compliance")[0]
    _, path = tempfile.mkstemp(suffix=".h5", prefix="afmformats_test_")

    with h5py.File(path, "w") as h5:
        fdat.export_data(h5, metadata=True, fmt="hdf5")
        fdat.export_data(h5, metadata=True, fmt="hdf5")

    with h5py.File(path, "r") as h5:
        assert h5["0"].attrs["enum"] == 0
        assert h5["1"].attrs["enum"] == 1
        assert h5["0"]["force"].attrs["unit"] == "N"


def test_load_custom_class():
    class TESTAFMData(afmformats.AFMCreepCompliance):
        pass
    jpkfile = data_path / "fmt-jpk-cc_pr14-brain-2021.06.30.jpk-force"
    fdat = afmformats.load_data(jpkfile,
                                modality="creep-compliance",
                                data_classes_by_modality={
                                    "creep-compliance": TESTAFMData
                                }
                                )[0]
    assert isinstance(fdat, TESTAFMData)


def test_segment():
    jpkfile = data_path / "fmt-jpk-cc_pr14-brain-2021.06.30.jpk-force"
    fdat = afmformats.load_data(jpkfile, modality="creep-compliance")[0]

    s1 = fdat.appr["segment"]
    assert np.all(s1 == 0)

    s2 = fdat.intr["segment"]
    assert np.all(s2 == 1)

    s2 = fdat.retr["segment"]
    assert np.all(s2 == 2)


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""Test of basic opening functionalities"""
import pathlib

import numpy as np

import afmformats
from afmformats.formats.fmt_jpk import JPKReader


data_path = pathlib.Path(__file__).resolve().parent / "data"


def test_meta_extracted_map():
    jpkfile = data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map"
    md = afmformats.load_data(jpkfile)[0].metadata
    summary = md.get_summary()
    for sec in ["acquisition", "dataset", "setup", "storage", "qmap"]:
        for kk in summary[sec]:
            v = summary[sec][kk]
            if not isinstance(v, (str, pathlib.Path)):
                assert not np.isnan(v), "should not be nan: '{}'".format(kk)


def test_meta_extracted_map2():
    jpkfile = data_path / "fmt-jpk-fd_map0d_extracted.jpk-force-map"
    md = afmformats.load_data(jpkfile)[0].metadata
    summary = md.get_summary()
    for sec in ["acquisition", "dataset", "setup", "storage", "qmap"]:
        for kk in summary[sec]:
            v = summary[sec][kk]
            if not isinstance(v, (str, pathlib.Path)):
                assert not np.isnan(v), "should not be nan: '{}'".format(kk)


def test_meta_extracted_single():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    md = afmformats.load_data(jpkfile)[0].metadata
    summary = md.get_summary()
    for sec in ["acquisition", "dataset", "setup", "storage"]:
        for kk in summary[sec]:
            v = summary[sec][kk]
            if not isinstance(v, (str, pathlib.Path)):
                assert not np.isnan(v), "should not be nan: '{}'".format(kk)


def test_meta_simple():
    jpkfilemap = data_path / "fmt-jpk-fd_map0d_extracted.jpk-force-map"
    jpkr = JPKReader(jpkfilemap)
    jpkr.get_metadata(index=0, segment=0)


def test_meta_single():
    jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
    jpkr = JPKReader(jpkfile)
    md = jpkr.get_metadata(index=0, segment=0)
    assert md["imaging mode"] == "force-distance"
    assert md["duration"] == 0.9999999999999998


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in list(_loc.keys()):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()
